# Authors: Fulvio Gesmundo, Leonie Kayser, Simon Telen
# Date: April 18, 2023
# Description: This file contains some examples of how to use Catalecticant.jl, a Julia package supporting the
# manuscript [GKT].

# Before starting, make sure that your working directory contains the folder Catalecticant and run
# using Pkg; Pkg.activate(Catalecticant)

# [GKT] Fulvio Gesmundo, Leonie Kayser, Simon Telen. "Hilbert Functions of Chopped Ideals". arXiv: 2307.02560

using Catalecticant
using DynamicPolynomials
using LinearAlgebra

# Define the parameters. The notation is as in [GKT]. We will decompose a degree D = 10 form
# in n+1 = 3 homogeneous variables y of rank r = 18.

n = 2
D = 10
r = 18
@polyvar y[1:n+1]

# We draw the coefficients of the linear forms from the uniform distribution on the complex unit circle,
# and expand the sum of the D-th powers of these linear forms to obtain f.

Z = exp.(2*pi*im*rand(r, n+1))
F = sum((Z*y).^D)

# We compute the waring decomposition f = ∑ cᵢ (zᵢ ̇ y)ᴰ of our form f in variables y.
# The output is a list of coefficients c, and a matrix "linforms" whose rows are zᵢ.

@time c, linforms = waring_decomposition(F,y,r)

# We compare the result with the original form f by comparing the coefficients.
g = dot(conj.(c),(linforms*y).^D)
println(norm(coefficients(g - F)./coefficients(F),Inf))

# Here is how to construct the catalecticant matrix from degree d to D-d:
d = 5 # make sure d < D
C = get_catalecticant(D-d,d,F)
println(rank(C)) # This should equal the rank

# For small enough ranks, the chopped ideal of the points zᵢ in f = ∑ cᵢ (zᵢ ̇ y)ᴰ is generated
# by the kernel of the middle catalecticant. In our example, we find three quintics.
d = Int(floor(D/2))
@polyvar x[1:n+1]
F = get_catalecticant_kernel(F,d,x)


# Our package allows to compute the conjectured smallest value of d' > d such that the Hilbert
# function of S/I<d>, with I<d> the chopped ideal of I(Z), equals r in degree d' (see Conjecture 1 in [GKT]).
# It is explained in [GKT, Section 3] that, in our example with d = 5, r = 18, n = 2, the answer is 7.
dreg = get_dreg(n,r,d)

# The size of the matrix built in the eigenvalue algorithm is the Hilbert function of the polynomial ring
# S in n+1 variables in degree dreg:
h_S(dreg,n)

# The size of this matrix is printed when running "waring_decomposition(f,x,r)", in the line "Sylv is a matrix of size ...".
