module Catalecticant

# Authors: Fulvio Gesmundo, Leonie Kayser, Simon Telen
# Date: April 18, 2023
# Description: This code accompanies the paper [GKT] entitled "Hilbert functions of chopped ideals".
#              It provides an implementation of the catalecticant method for Waring decomposition in the low rank regime.
#              For the polynomial system solving step, we use the implementation in [BT] for solving equations using eigenvalues.
#
# [BT] Matías R Bender and Simon Telen. "Yet another eigenvalue algorithm for solving polynomial systems". In: arXiv preprint arXiv:2105.08472 (2021)
# [GKT] Fulvio Gesmundo, Leonie Kayser, Simon Telen. "Hilbert functions of chopped ideals". In preparation.

export deriv, get_catalecticant, h_S, get_dreg, get_catalecticant_kernel, get_admissible_tuple, get_coefficients, waring_decomposition

include("JuliaEigenvalueSolver-main/src/EigenvalueSolver.jl")
using DynamicPolynomials
using LinearAlgebra

# Compute the derivative (∂^|v| f)/(∂x^v) of a polynomial f.
function deriv(f,v,x)
    vars = x
    n = length(v)
    g = copy(f)
    for i = 1:n
        g = differentiate(g,vars[i],v[i])
    end
    return g
end

# Compute the catalecticant matrix of a homogeneous polynomial f which maps polynomials of degree v to "dual" polynomials of degree u.
# it is assumed that f is homogeneous and u + v = deg(f).
function get_catalecticant(u,v,f)
    vars = variables(f)
    mons_v = monomials((sum(vars))^v)
    exps_v = exponents.(mons_v)
    mons_u = monomials((sum(vars))^u)
    C = zeros(length(mons_u),length(mons_v)) + im*zeros(length(mons_u),length(mons_v))
    for i = 1:length(mons_v)
        g = deriv(f,exps_v[i],vars)
        C[:,i] = coefficients(g)
    end
    return C
end

# Return the Hilbert function of Pⁿ in degree d
function h_S(d,n)
    if d < 0
        return 0
    else
        return binomial(d+n,n)
    end
end

# For a chopped ideal I<d> generated by degree d forms defining r points in Pⁿ,
# this function predicts the first degree dd > d such that h_{S/I<d>} = r.
function get_dreg(n,r,d)
    if r <= binomial(n+d-1,n) || r >= binomial(n+d,n)-n
        println("r is not in the right range")
    else
        e = 1
        while h_S(d+e,n)-r > sum([(-1)^k*h_S(e-k*d,n)*binomial(h_S(d,n)-r,k+1) for k = 0:n-2])
            e += 1
        end
        return d + e
    end
end

# Compute a basis for the catalecticant kernel of a homogeneous polynomial f.
# The basis elements are polynomials of degree d in the variables x.
# It is assumed that d < deg(f).
function get_catalecticant_kernel(f,d,x)
    D = degree(monomials(f)[1])
    v = d
    u = D-d
    C = get_catalecticant(u,v,f)
    mons_v = monomials(sum(x)^v)
    N = nullspace(C)
    F = transpose(N)*mons_v
    return F
end

# Compute an admissible tuple in the sense of [BT21] which allows to compute the points
# defined by the chopped ideal of degree d.
function get_admissible_tuple(X,dreg,d,s)
    Dmons = EigenvalueSolver.getMonomials(X, dreg)
    Emons = [EigenvalueSolver.getMonomials(X, dreg-d) for i = 1:s]
    Emons = [[EigenvalueSolver.getMonomials(X, d)]; Emons]
    A₀mons = EigenvalueSolver.getMonomials(X, dreg-d)
    D = EigenvalueSolver.montoexp(Dmons)
    E = [EigenvalueSolver.montoexp(EE) for EE ∈ Emons]
    A₀ = EigenvalueSolver.montoexp(A₀mons)
    return D, E, A₀
end

# Recover the Waring decomposition f = ∑ cᵢ (zᵢ ̇ x)ᴰ from the rows zᵢ of linforms.
function get_coefficients(f,linforms,x; safety_factor = 1.5)
    nsamples = Int(ceil(size(linforms,1)*safety_factor))
    samples = [exp.(2*pi*im*rand(length(x))) for i = 1:nsamples]
    sample_mtx = hcat(samples...)
    D = degree(monomials(f)[1])
    A = transpose((linforms*sample_mtx).^D)
    b = [convert(ComplexF64,subs(f,x=>sample_mtx[:,i])) for i = 1:nsamples]
    c = A\b
    println(norm(A*c-b))
    return c
end

# Compute the Waring decomposition f = ∑ cᵢ (zᵢ ̇ x)ᴰ of a degree D form f in variables x of rank r.
# The output is a list of coefficients c, and a matrix "linforms" whose rows are zᵢ.
# To compare the result to the original polynomial, use for instance
# julia> g = dot(conj.(c),(linforms*x).^D)
# julia> println(coefficients(g - f))
function waring_decomposition(f,x,r)
    D = degree(monomials(f)[1])
    d = Int(floor(D/2))
    n = length(x)-1
    F = get_catalecticant_kernel(f,d,x)
    F = subs(F,x[end]=>1)
    dreg = get_dreg(n,r,d)
    X = x[1:end-1]
    DD, E, A₀ = get_admissible_tuple(X,dreg,d,length(F))
    sol = EigenvalueSolver.solve_EV(F,X,A₀,E,DD; rankTol = 1e3 * eps(), schurfac = true, check_criterion = false, compress = false, verbose = true)
    linforms = transpose(hcat([[s;1] for s in sol]...))
    c = get_coefficients(f,linforms,x)
    return c, linforms
end


end # module Catalecticant
